__all__ = [
    'AbstractDocumentContent', 'AbstractTreeDocumentContent', 'NodeType',
    'AbstractTalismanDocument',
    'AbstractDirective',
    'AbstractFact', 'FactStatus', 'FactType',
    'AbstractFact', 'FactStatus', 'FactMetadata', 'FactType',
    'AbstractSpan', 'AbstractTalismanSpan'
]

from .content import AbstractDocumentContent, AbstractTreeDocumentContent, NodeType
from .directive import AbstractDirective
from .document import AbstractTalismanDocument
from .fact import AbstractFact, FactMetadata, FactStatus, FactType
from .span import AbstractSpan, AbstractTalismanSpan



from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, Generic, Iterable, Iterator, Optional, Tuple, Type, TypeVar

from .content import AbstractTreeDocumentContent
from .directive import AbstractDirective
from .fact import AbstractFact

DocumentMetadata = Dict[str, Any]
_AbstractTalismanDocument = TypeVar('_AbstractTalismanDocument', bound='AbstractTalismanDocument')
_AbstractTreeDocumentContent = TypeVar('_AbstractTreeDocumentContent', bound=AbstractTreeDocumentContent)
_Fact = TypeVar('_Fact', bound=AbstractFact)
_Directive = TypeVar('_Directive', bound=AbstractDirective)


class AbstractTalismanDocument(Generic[_AbstractTreeDocumentContent], metaclass=ABCMeta):
    __slots__ = ()

    @property
    @abstractmethod
    def doc_id(self) -> str:
        pass

    @property
    @abstractmethod
    def content(self) -> _AbstractTreeDocumentContent:
        pass

    @abstractmethod
    def with_content(self: _AbstractTalismanDocument, content: _AbstractTreeDocumentContent) -> _AbstractTalismanDocument:
        pass

    @property
    @abstractmethod
    def metadata(self) -> Optional[DocumentMetadata]:
        pass

    @abstractmethod
    def with_metadata(self: _AbstractTalismanDocument, metadata: DocumentMetadata) -> _AbstractTalismanDocument:
        pass

    @property
    @abstractmethod
    def facts(self) -> Tuple[AbstractFact, ...]:
        pass

    @abstractmethod
    def filter_facts(self, type_: Type[_Fact], filter_: Callable[[_Fact], bool] = lambda _: True) -> Iterator[_Fact]:
        pass

    @abstractmethod
    def with_facts(self: _AbstractTalismanDocument, facts: Iterable[AbstractFact]) -> _AbstractTalismanDocument:
        pass

    @abstractmethod
    def without_facts(self: _AbstractTalismanDocument) -> _AbstractTalismanDocument:
        pass

    @property
    @abstractmethod
    def directives(self) -> Tuple[AbstractDirective, ...]:
        pass

    @abstractmethod
    def with_directives(self: _AbstractTalismanDocument, directives: Iterable[AbstractDirective]) -> _AbstractTalismanDocument:
        pass

    @abstractmethod
    def without_directives(self: _AbstractTalismanDocument) -> _AbstractTalismanDocument:
        pass

    @abstractmethod
    def filter_directives(self, type_: Type[_Directive], filter_: Callable[[_Directive], bool] = lambda _: True) -> Iterator[_Directive]:
        pass

    @abstractmethod
    def fact_text_mention(self, fact: AbstractFact) -> Optional[Tuple[str, ...]]:
        pass



__all__ = [
    'NodeMarkup', 'NodeMetadata', 'TreeDocumentContentModel',
    'DocumentMetadataModel', 'TalismanDocumentModel',
    'ConceptFactModel', 'PropertyFactModel', 'PropertyLinkValueModel', 'RelationFactModel', 'SpanModel', 'ValueFactModel',
    'build_fact_model',
    'FactMetadataModel'
]

from .content import NodeMarkup, NodeMetadata, TreeDocumentContentModel
from .document import DocumentMetadataModel, TalismanDocumentModel
from .fact import ConceptFactModel, PropertyFactModel, PropertyLinkValueModel, RelationFactModel, SpanModel, ValueFactModel, \
    build_fact_model
from .fact.common import FactMetadataModel



from abc import abstractmethod
from typing import Type, TypeVar

from pydantic import BaseModel

from tdm.abstract.datamodel.directive import AbstractDirective, DirectiveType

_AbstractDirectiveModel = TypeVar('_AbstractDirectiveModel', bound='AbstractDirectiveModel')


class AbstractDirectiveModel(BaseModel):
    directive_type: DirectiveType

    @classmethod
    @abstractmethod
    def build(cls: Type[_AbstractDirectiveModel], directive: AbstractDirective) -> _AbstractDirectiveModel:
        pass

    @abstractmethod
    def to_directive(self) -> AbstractDirective:
        pass



from collections import defaultdict
from typing import Dict, Optional, Tuple, Union

from pydantic import BaseModel

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanDocument, FactType
from tdm.datamodel import TalismanDocument
from .content import TreeDocumentContentModel
from .directive import CreateAccountDirectiveModel, CreateConceptDirectiveModel, CreatePlatformDirectiveModel, build_directive_model
from .fact import ConceptFactModel, PropertyFactModel, RelationFactModel, ValueFactModel, build_fact_model
from ..abstract.json_schema import MetadataModel

FactModel = Union[ConceptFactModel, ValueFactModel, PropertyFactModel, RelationFactModel]
DirectiveModel = Union[CreateAccountDirectiveModel, CreateConceptDirectiveModel, CreatePlatformDirectiveModel]


class DocumentMetadataModel(MetadataModel):
    title: Optional[str]
    file_name: Optional[str]
    file_type: Optional[str]
    size: Optional[int]
    created_time: Optional[int]
    access_time: Optional[int]
    modified_time: Optional[int]
    publication_date: Optional[int]
    publication_author: Optional[str]
    description: Optional[str]
    language: Optional[str]
    parent_uuid: Optional[str]
    url: Optional[str]
    platform: Optional[str]
    account: Optional[Tuple[str, ...]]
    access_level: Optional[str]
    user: Optional[str]
    path: Optional[str]
    trust_level: Optional[float]
    markers: Optional[Tuple[str, ...]]
    related_concept_id: Optional[str]
    preview_text: Optional[str]
    story: Optional[str]

    class Config:
        extra = 'allow'  # any other extra fields will be kept


class TalismanDocumentModel(BaseModel):
    id: str
    content: TreeDocumentContentModel
    metadata: Optional[DocumentMetadataModel]
    facts: Optional[Tuple[FactModel, ...]]  # pydantic fails to serialize FrozenSet[FactModel]
    directives: Optional[Tuple[DirectiveModel, ...]]

    def to_doc(self) -> TalismanDocument:
        type2fact_factory = defaultdict(list)
        for t, ff in map(lambda model: model.to_fact_factory(), (self.facts or tuple())):
            type2fact_factory[t].append(ff)

        def build_facts(fact_type: FactType, ids_mapping: Dict[str, AbstractFact]) -> Dict[str, AbstractFact]:
            return {fact.id: fact for fact in map(lambda factory: factory(ids_mapping), type2fact_factory.get(fact_type, []))}
        ids2facts = {}
        for ft in [FactType.CONCEPT, FactType.VALUE, FactType.RELATION, FactType.PROPERTY]:
            ids2facts.update(build_facts(ft, ids2facts))

        return TalismanDocument(
            doc_id=self.id, content=self.content.to_content(),
            metadata=self.metadata.to_metadata() if self.metadata is not None else None,
            facts=tuple(ids2facts.values()),
            directives=tuple(directive.to_directive() for directive in self.directives) if self.directives else None
        )

    @classmethod
    def build(cls, doc: AbstractTalismanDocument) -> 'TalismanDocumentModel':
        content_type = cls.__fields__['content'].type_
        return cls.construct(
            id=doc.doc_id,
            content=content_type.build(doc.content),
            metadata=DocumentMetadataModel(**doc.metadata) if doc.metadata is not None else None,
            facts=tuple(build_fact_model(fact, doc.content) for fact in doc.facts) if doc.facts is not None else None,
            directives=tuple(build_directive_model(directive) for directive in doc.directives) if doc.directives else None
        )



from abc import ABCMeta, abstractmethod
from typing import Iterable, Optional, Tuple, TypeVar

_AbstractTree = TypeVar('_AbstractTree', bound='AbstractTree')


class AbstractTree(metaclass=ABCMeta):
    __slots__ = ()

    @property
    @abstractmethod
    def id(self) -> str:
        pass

    @property
    def root(self: _AbstractTree) -> _AbstractTree:
        parent = self.parent
        if parent is None:
            return self
        return parent.root

    @property
    @abstractmethod
    def parent(self: _AbstractTree) -> Optional[_AbstractTree]:
        pass

    @property
    @abstractmethod
    def nodes(self: _AbstractTree) -> Tuple[_AbstractTree, ...]:
        pass

    @abstractmethod
    def node(self: _AbstractTree, node_id: str) -> _AbstractTree:
        pass

    @abstractmethod
    def detached_node(self: _AbstractTree, node_id: str) -> _AbstractTree:
        pass

    @abstractmethod
    def contain_node(self: _AbstractTree, node_id: str) -> bool:
        pass

    @abstractmethod
    def pruned(self: _AbstractTree) -> _AbstractTree:
        pass

    @abstractmethod
    def with_nodes(self: _AbstractTree, nodes: Iterable[_AbstractTree]) -> _AbstractTree:
        pass

    @abstractmethod
    def as_root(self: _AbstractTree) -> _AbstractTree:
        pass

    @abstractmethod
    def equal_structure(self: _AbstractTree, other: _AbstractTree) -> bool:
        pass

    @abstractmethod
    def _get_updated_constructor_params(self, **kwargs) -> dict:
        pass

    def _with_changes(self: _AbstractTree, **kwargs) -> _AbstractTree:
        return type(self)(**self._get_updated_constructor_params(**kwargs))



from abc import abstractmethod
from typing import Any, Dict, Type, TypeVar

from pydantic import BaseModel

from .datamodel import AbstractTreeDocumentContent

_AbstractContentModel = TypeVar('_AbstractContentModel', bound='AbstractContentModel')


class AbstractContentModel(BaseModel):
    @abstractmethod
    def to_content(self) -> AbstractTreeDocumentContent:
        pass

    @classmethod
    @abstractmethod
    def build(cls: Type[_AbstractContentModel], doc: AbstractTreeDocumentContent) -> _AbstractContentModel:
        pass


class MetadataModel(BaseModel):
    def to_metadata(self) -> Dict[str, Any]:
        return self.dict(exclude_none=True)
